@@ -19,11 +19,13 @@
 
 #include <config.h>
 
-#include "debug.h"
+#include "debug.hh"
 #include "vterowdata.hh"
 
 #include <string.h>
 
+#include <algorithm>
+#include <iterator>
 #include <type_traits>
 
 /* This will be true now that VteCell is POD, but make sure it'll be true
@@ -51,7 +53,7 @@ _vte_cells_for_cell_array (VteCell *cells)
 	if (G_UNLIKELY (!cells))
 		return NULL;
 
-	return (VteCells *) (((guchar *) cells) - G_STRUCT_OFFSET (VteCells, cells));
+	return reinterpret_cast<VteCells*>(((guchar *) cells) - G_STRUCT_OFFSET (VteCells, cells));
 }
 
 static VteCells *
@@ -59,7 +61,9 @@ _vte_cells_realloc (VteCells *cells, guint32 len)
 {
 	guint32 alloc_len = (1 << g_bit_storage (MAX (len, 80))) - 1;
 
-	_vte_debug_print(VTE_DEBUG_RING, "Enlarging cell array of %d cells to %d cells\n", cells ? cells->alloc_len : 0, alloc_len);
+	_vte_debug_print(vte::debug::category::RING,
+                         "Enlarging cell array of {} cells to {} cells",
+                         cells ? cells->alloc_len : 0, alloc_len);
 	cells = (VteCells *)g_realloc (cells, G_STRUCT_OFFSET (VteCells, cells) + alloc_len * sizeof (cells->cells[0]));
 	cells->alloc_len = alloc_len;
 
@@ -69,7 +73,9 @@ _vte_cells_realloc (VteCells *cells, guint32 len)
 static void
 _vte_cells_free (VteCells *cells)
 {
-	_vte_debug_print(VTE_DEBUG_RING, "Freeing cell array of %d cells\n", cells->alloc_len);
+	_vte_debug_print(vte::debug::category::RING,
+                         "Freeing cell array of {} cells",
+                         cells->alloc_len);
 	g_free (cells);
 }
 
@@ -100,7 +106,7 @@ _vte_row_data_fini (VteRowData *row)
 	row->cells = NULL;
 }
 
-static inline gboolean
+static inline bool
 _vte_row_data_ensure (VteRowData *row, gulong len)
 {
 	VteCells *cells = _vte_cells_for_cell_array (row->cells);
@@ -115,6 +121,12 @@ _vte_row_data_ensure (VteRowData *row, gulong len)
 	return TRUE;
 }
 
+bool
+_vte_row_data_ensure_len (VteRowData *row, gulong len)
+{
+        return _vte_row_data_ensure(row, len);
+}
+
 void
 _vte_row_data_insert (VteRowData *row, gulong col, const VteCell *cell)
 {
@@ -153,13 +165,21 @@ void _vte_row_data_remove (VteRowData *row, gulong col)
 void _vte_row_data_fill (VteRowData *row, const VteCell *cell, gulong len)
 {
 	if (row->len < len) {
-		gulong i;
-
 		if (G_UNLIKELY (!_vte_row_data_ensure (row, len)))
 			return;
 
-		for (i = row->len; i < len; i++)
-			row->cells[i] = *cell;
+		std::fill_n(&row->cells[row->len], len - row->len, *cell);
+		row->len = len;
+	}
+}
+
+/* This leaves the new cells uninitialized, potentially containing random data.
+ * It's the caller's responsibility to initialize them. */
+void _vte_row_data_expand (VteRowData *row, gulong len)
+{
+	if (row->len < len) {
+		if (G_UNLIKELY (!_vte_row_data_ensure (row, len)))
+			return;
 
 		row->len = len;
 	}
@@ -179,6 +199,25 @@ void _vte_row_data_copy (const VteRowData *src, VteRowData *dst)
         memcpy(dst->cells, src->cells, src->len * sizeof (src->cells[0]));
 }
 
+void _vte_row_data_fill_cells(VteRowData* row,
+                              gulong start_idx,
+                              VteCell const* fill_cell,
+                              VteCell const* cells,
+                              gulong len)
+{
+        auto const needlen = start_idx + len;
+        if (!_vte_row_data_ensure(row, needlen))
+                return;
+
+        // Fill up to start_idx with @fill_cell
+        _vte_row_data_fill(row, fill_cell, start_idx);
+        // ... then copy the cells over ...
+        std::copy_n(cells, len, &row->cells[start_idx]);
+        // ... and adjust the row length
+        if (row->len < needlen)
+                row->len = needlen;
+}
+
 /* Get the length, ignoring trailing empty cells (with a custom background color). */
 guint16 _vte_row_data_nonempty_length (const VteRowData *row)
 {
